/* VMS__CVT.C -- 14-APR-1996 Uwe Zessin

   19-AUG-1997 ZE. -- vms__cvt_u2v_time()
   05-NOV-1997 ZE. -- vms__cvt_oct2hex() + vms__cvt_hex2oct()
   14-DEC-1997 ZE. -- include LIB$ROUTINES.H,
			add vms__cvt_v2u_time() + vms__cvt_test_t()
*/

/* conversion routines */

#include <descrip.h>		/* string descriptor support */
#include <lib$routines.h>	/* LIB$name */
#include <ots$routines.h>	/* OTS$name */
#include <ssdef.h>		/* SS$_name */
#include <stdio.h>		/* print() */
#include <time.h>

/* ------------------------------------------------------------------------- */
/* input a binary quadword, output a hex string */
/* you MUST provide storage for 16 characters at *t_quadword_hex ! */
long vms__cvt_quad2hex (long* q_quadword_bin, char* t_quadword_hex)
{
	char*			at_chr;
	long			l_longword;
	struct dsc$descriptor_s r_res_str;
	long			l_status;

	/* initialize hex-string to all '?' */
	at_chr = t_quadword_hex;
	/* mis-use l_longword as loop counter */
	for (l_longword = 0; l_longword < 16; l_longword++)
	{
	    *at_chr = '?';
	    at_chr++;			/* next character */
	}

	/* convert first (low) longword to hex-string */
	l_longword = *q_quadword_bin;		  /* value of 1st longword */

	r_res_str.dsc$w_length  = 8;		  /* longword hex string */
	r_res_str.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class   = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = (t_quadword_hex + 8); /* low-longword */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);
	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* skip second conversion */
	}

	/* convert second (high) longword to hex-string */
	l_longword = *(q_quadword_bin + 1);	/* value of 2nd longword */
					/* 1 = next INTEGER (= 4 byte)! */

	r_res_str.dsc$a_pointer = t_quadword_hex; /* second longword string */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);

	return l_status;
} /* vms__cvt_quad2hex() */

/* ------------------------------------------------------------------------- */
/* input a hex string, output a binary quadword */
/* you MUST provide 16 hex characters at *t_quadword_hex ! */
long vms__cvt_hex2quad (char* t_quadword_hex, long* q_quadword_bin)
{
	char*			t_longword;
	long			l_longword;
	struct dsc$descriptor_s r_res_str;
	long			l_status;

	/* convert first part of hex-string to higher longword */
	t_longword = t_quadword_hex;

	r_res_str.dsc$w_length = 8;		/* a longword */
	r_res_str.dsc$b_dtype = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = t_longword;	/* first longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;			/* skip second conversion */
	}

	*(q_quadword_bin + 1) = l_longword;	/* store higher longword */
			/* 1 = next INTEGER (= 4 byte)! */
	
	/* convert second part of hex-string to lower longword */
	t_longword = t_quadword_hex + 8;
	r_res_str.dsc$a_pointer = t_longword;	/* second longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* error while second conversion */
	}

	*q_quadword_bin = l_longword;	/* store higher longword */

	return SS$_NORMAL;
} /* vms__cvt_hex2quad() */

/* ------------------------------------------------------------------------- */
/* test example for hex/bin conversion */
long vms__cvt_test_q (void)
{
	char	t_hex0[] = "----------------";
	char	t_hex1[] = "12345678ABCDEF90";
	long	q_quad0[2];
	long	q_quad1[2];
	char*	cp;
	long	l_status;

	q_quad0[0] = 0x12345678;
	q_quad0[1] = 0xabcdef90;
	cp = &t_hex0[0];
	l_status = vms__cvt_quad2hex (&q_quad0[0], cp);
	/* DBG> examine /hex q_quad0[0], q_quad0[1] */

	cp = &t_hex1[0];
	l_status = vms__cvt_hex2quad (cp, &q_quad1[0]);
	/* DBG> examine /asciz *cp */

	return SS$_NORMAL;
} /* vms__cvt_test_q() */

/* ------------------------------------------------------------------------- */
/* input a binary octaword, output a hex string */
/* you MUST provide storage for 32 characters at *t_octaword_hex ! */
long vms__cvt_oct2hex (long* o_octaword_bin, char* t_octaword_hex)
{
	char*			at_chr;
	long			l_longword;
	struct dsc$descriptor_s r_res_str;
	long			l_status;

	/* initialize hex-string to all '?' */
	at_chr = t_octaword_hex;
	/* mis-use l_longword as loop counter */
	for (l_longword = 0; l_longword < 32; l_longword++)
	{
	    *at_chr = '?';
	    at_chr++;			/* next character */
	}

	/* convert first longword to hex-string */
	l_longword = *o_octaword_bin;		  /* value of 1st longword */

	r_res_str.dsc$w_length  = 8;		  /* longword hex string */
	r_res_str.dsc$b_dtype   = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class   = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = (t_octaword_hex + 24); /* low-longword */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);
	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* skip second conversion */
	}

	/* convert second longword to hex-string */
	l_longword = *(o_octaword_bin + 1);	/* value of 2nd longword */
					/* 1 = next INTEGER (= 4 byte)! */

	r_res_str.dsc$a_pointer = (t_octaword_hex + 16); /* second longword string */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);

	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* skip third conversion */
	}

	/* convert third longword to hex-string */
	l_longword = *(o_octaword_bin + 2);	/* value of 3rd longword */
					/* 1 = next INTEGER (= 4 byte)! */

	r_res_str.dsc$a_pointer = (t_octaword_hex +8); /* third longword string */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);

	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* skip fourth conversion */
	}

	/* convert fourth longword to hex-string */
	l_longword = *(o_octaword_bin + 3);	/* value of 4th longword */
					/* 1 = next INTEGER (= 4 byte)! */

	r_res_str.dsc$a_pointer = t_octaword_hex; /* fourth longword string */

	l_status = ots$cvt_l_tz		/* convert longword to hex string */
		(&l_longword		/* varying-input-value		  */
		,&r_res_str		/* fixed-length-resultant-string  */
		,8			/* number-of-digits		  */
		,4			/* input-value-size		  */
		);

	return l_status;

} /* vms__cvt_oct2hex() */

/* ------------------------------------------------------------------------- */
/* input a hex string, output a binary octaword */
/* you MUST provide 32 hex characters at *t_octaword_hex ! */
long vms__cvt_hex2oct (char* t_octaword_hex, long* o_octaword_bin)
{
	char*			t_longword;
	long			l_longword;
	struct dsc$descriptor_s r_res_str;
	long			l_status;

	/* convert first part of hex-string to fourth longword */
	t_longword = t_octaword_hex;

	r_res_str.dsc$w_length = 8;		/* a longword */
	r_res_str.dsc$b_dtype = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = t_longword;	/* first longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;			/* skip second conversion */
	}

	*(o_octaword_bin + 3) = l_longword;	/* store 4th longword */
			/* 1 = next INTEGER (= 4 byte)! */

	/* convert second part of hex-string to third longword */
	t_longword = t_octaword_hex + 8;

	r_res_str.dsc$w_length = 8;		/* a longword */
	r_res_str.dsc$b_dtype = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = t_longword;	/* second longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;			/* skip third conversion */
	}

	*(o_octaword_bin + 2) = l_longword;	/* store third longword */
			/* 1 = next INTEGER (= 4 byte)! */

	/* convert third part of hex-string to second longword */
	t_longword = t_octaword_hex + 16;

	r_res_str.dsc$w_length = 8;		/* a longword */
	r_res_str.dsc$b_dtype = DSC$K_DTYPE_T;
	r_res_str.dsc$b_class = DSC$K_CLASS_S;
	r_res_str.dsc$a_pointer = t_longword;	/* third longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;			/* skip fourth conversion */
	}

	*(o_octaword_bin + 1) = l_longword;	/* store third longword */
			/* 1 = next INTEGER (= 4 byte)! */

	/* convert fourth part of hex-string to first longword */
	t_longword = t_octaword_hex + 24;
	r_res_str.dsc$a_pointer = t_longword;	/* fourth longword string */

	l_status = ots$cvt_tz_l		/* convert hex string to longword */
		(&r_res_str		/* fixed-or-dynamic-input-string  */
		,&l_longword	  	/* varying-output-value		  */
		,4			/* [,output-value-size]		  */
		,0			/* [,flags-value]		  */
		);			/*  0=interpret blanks as zeroes  */
	if (l_status != SS$_NORMAL)
	{
	    return l_status;		/* error while fourth conversion */
	}

	*o_octaword_bin = l_longword;	/* store fourth longword */

	return SS$_NORMAL;
} /* vms__cvt_hex2oct() */

/* ------------------------------------------------------------------------- */
/* test example for hex/bin conversion */
/* DBG> set module vms__cvt	   */
/* DBG> set break vms__cvt_test_o  */
/* DBG> call vms__cvt_test_o	   */
long vms__cvt_test_o (void)
{
	char	t_hex0[] = "--------------------------------";
	char	t_hex1[] = "ABCDEF901234567843218765EF90ABCD";
	long	q_oct0[4];
	long	q_oct1[4];
	char*	cp;
	long	l_status;

	q_oct0[0] = 0x12345678;
	q_oct0[1] = 0xabcdef90;
	q_oct0[2] = 0x87654321;
	q_oct0[3] = 0x09fedcba;
	cp = &t_hex0[0];
	l_status = vms__cvt_oct2hex (&q_oct0[0], cp);
	/* DBG> examine /hex q_oct0[0], q_oct0[1], q_oct0[2], q_oct0[3] */

	cp = &t_hex1[0];
	l_status = vms__cvt_hex2oct (cp, &q_oct1[0]);
	/* DBG> examine /asciz *cp */

	return SS$_NORMAL;
} /* vms__cvt_test_o() */

/* ------------------------------------------------------------------------- */
/* convert Unix time_t value to VMS bintim value */
void vms__cvt_u2v_time (time_t al_unixtime, unsigned int * aq_vmstime[2])
		     /* input             , output                       */
{
    long l_addend = 0;		/* -> lib$emul() */
    long l_arraylength = 2;	/* -> lib$addx() */

    /* time_t l_unixtime; */	/* time_t tick is 1 second */
    long l_multiplicant = 10000000L;
				/* VMS bintim tick is 100ns (nano-second) */
				/* 1 second = 10000000 * 100ns */

    unsigned int q_basetime[2];	/* 01-JAN-1970 00:00:00.00 in bintim format */
    unsigned int q_addstime[2];	/* add seconds in bintim format */

    q_basetime[0] = 0x4BEB4000L;	/* 01-JAN-1970 00:00:00.00 */
    q_basetime[1] = 0x007C9567L;	/* in VMS bintim format    */

    (void) lib$emul
	(&al_unixtime		/* longword-integer-multiplier   */
	,&l_multiplicant	/* longword-integer-multiplicand */
	,&l_addend		/* addend  */
	,&q_addstime[0]		/* product */
	);

    /* ignore condition value SS$_INTOVF */
    (void) lib$addx
	(&q_basetime[0]		/* addend-array		     */
	,&q_addstime[0]		/* augend-array		     */
	,aq_vmstime		/* resultant-array	     */
	,&l_arraylength		/* [,array-length] longwords */
	);
} /* vms__cvt_u2v_time() */

/* ------------------------------------------------------------------------- */
/* convert VMS bintim value to Unix time_t value */
void vms__cvt_v2u_time (unsigned int * aq_vmstime, time_t * al_unixtime)
		     /* input			    , output		*/
{
    long l_remainder   = 0;	/* -> lib$ediv() */
    long l_arraylength = 2;	/* -> lib$addx() */

    /* time_t l_unixtime; */	/* time_t tick is 1 second */
    long l_divisor = 10000000L;	/* VMS bintim tick is 100ns (nano-second) */
				/* 1 second = 10000000 * 100ns */

    unsigned int q_basetime[2];	/* 01-JAN-1970 00:00:00.00 in bintim format */
    unsigned int q_substime[2];	/* substract seconds in bintim format */

    q_basetime[0] = 0x4BEB4000L;	/* 01-JAN-1970 00:00:00.00 */
    q_basetime[1] = 0x007C9567L;	/* in VMS bintim format    */

    /* ignore condition value SS$_INTOVF */
    (void) lib$subx
	(aq_vmstime		/* minuend-array	     */
	,&q_basetime[0]		/* subtrahend-array	     */
	,&q_substime[0]		/* difference-array	     */
	,&l_arraylength		/* [,array-length] longwords */
	);

    (void) lib$ediv
	(&l_divisor		/* longword-integer-divisor	*/
	,&q_substime[0]		/* quadword-integer-dividend	*/
	,al_unixtime		/* longword-integer-quotient	*/
	,&l_remainder		/* remainder	*/
	);

} /* vms__cvt_v2u_time() */

/* ------------------------------------------------------------------------- */
/* test example for bintim<->time_t conversion */
/* DBG> set module vms__cvt	   */
/* DBG> set break vms__cvt_test_t  */
/* DBG> call vms__cvt_test_t	   */
long vms__cvt_test_t (void)
{
unsigned int q_testtime[2];	/* 29-FEB-2000 12:34:56.78 in bintim format */

time_t	l_testtime;		/* convert */
long	l_status;

q_testtime[0] = 0xD9755CC0L;	/* 29-FEB-2000 12:34:56.78 */
q_testtime[1] = 0x009E6639L;	/* in VMS bintim format    */

(void) vms__cvt_v2u_time (&q_testtime[0], &l_testtime);
(void) printf ("%s\n", ctime(&l_testtime));

	return SS$_NORMAL;
} /* vms__cvt_test_t() */

/* ------------------------------------------------------------------------- */

/* EOF: VMS__CVT.C */
